/*
 * Configurable ps-like program.
 * Command line option parsing.
 *
 * Copyright (c) 1995 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include "ips.h"


/*
 * Definitions for version printout.
 */
#define	VERSION		"1.4"
#define	AUTHOR_NAME	"David I. Bell"
#define	AUTHOR_EMAIL	"dbell@auug.org.au"


/*
 * Command line options.
 */
#define	OPT_ILLEGAL		0
#define	OPT_SET_COLUMNS		1
#define	OPT_ADD_COLUMNS		2
#define	OPT_REMOVE_COLUMNS	3
#define	OPT_SEPARATION		4
#define	OPT_WIDTH		5
#define	OPT_SLEEP_TIME		6
#define	OPT_NO_SELF		7
#define	OPT_MY_PROCS		9
#define	OPT_ACTIVE_ONLY		10
#define	OPT_NO_ROOT		11
#define	OPT_CLEAR_SCREEN	12
#define	OPT_READ_FILE		13
#define	OPT_NORMAL_SORT		14
#define	OPT_REVERSE_SORT	15
#define	OPT_NORMAL_SORT_EXPR	16
#define	OPT_NO_SORT		17
#define	OPT_LOOP		18
#define	OPT_INIT_TIME		19
#define	OPT_CONDITION		20
#define	OPT_TOP_COUNT		21
#define	OPT_NO_HEADER		22
#define	OPT_PIDS		23
#define	OPT_USERS		24
#define	OPT_HELP		25
#define	OPT_LIST_COLUMNS	26
#define	OPT_END			27
#define	OPT_NO_INIT		28
#define	OPT_ACTIVE_TIME		29
#define	OPT_LIST_MACROS		30
#define	OPT_VERSION		31
#define	OPT_DEFAULT		32
#define	OPT_SYNC_TIME		33
#define	OPT_GROUPS		34
#define	OPT_VERTICAL		35
#define	OPT_REVERSE_SORT_EXPR	36
#define	OPT_COLUMN_WIDTH	37
#define	OPT_LAST_OPTION		37


/*
 * Table of mappings between option names and option ids.
 */
typedef	struct	{
	char *	name;			/* option name */
	int	id;			/* option id */
} OPTION;


static	OPTION	option_table[] =
{
	{"col",		OPT_SET_COLUMNS},
	{"addcol",	OPT_ADD_COLUMNS},
	{"remcol",	OPT_REMOVE_COLUMNS},
	{"sep",		OPT_SEPARATION},
	{"width",	OPT_WIDTH},
	{"sleep",	OPT_SLEEP_TIME},
	{"noself",	OPT_NO_SELF},
	{"my",		OPT_MY_PROCS},
	{"active",	OPT_ACTIVE_ONLY},
	{"noroot",	OPT_NO_ROOT},
	{"clear",	OPT_CLEAR_SCREEN},
	{"read",	OPT_READ_FILE},
	{"sort",	OPT_NORMAL_SORT},
	{"revsort",	OPT_REVERSE_SORT},
	{"sortexpr",	OPT_NORMAL_SORT_EXPR},
	{"revsortexpr",	OPT_REVERSE_SORT_EXPR},
	{"nosort",	OPT_NO_SORT},
	{"loop",	OPT_LOOP},
	{"vert",	OPT_VERTICAL},
	{"initsleep",	OPT_INIT_TIME},
	{"cond",	OPT_CONDITION},
	{"top",		OPT_TOP_COUNT},
	{"noheader",	OPT_NO_HEADER},
	{"pid",		OPT_PIDS},
	{"user",	OPT_USERS},
	{"group",	OPT_GROUPS},
	{"h",		OPT_HELP},
	{"help",	OPT_HELP},
	{"listcolumns",	OPT_LIST_COLUMNS},
	{"end",		OPT_END},
	{"noinit",	OPT_NO_INIT},
	{"activetime",	OPT_ACTIVE_TIME},
	{"listmacros",	OPT_LIST_MACROS},
	{"version",	OPT_VERSION},
	{"default",	OPT_DEFAULT},
	{"synctime",	OPT_SYNC_TIME},
	{"colwidth",	OPT_COLUMN_WIDTH},
	{NULL,		OPT_ILLEGAL}
};


/*
 * Static routines.
 */
static	BOOL	ReadFile(ARGS *ap);
static	BOOL	SetSleepTime(ARGS *ap);
static	BOOL	SetActiveTime(ARGS *ap);
static	BOOL	SetSyncTime(ARGS *ap);
static	BOOL	SetTopCount(ARGS *ap);
static	BOOL	SetInitTime(ARGS *ap);
static	BOOL	SetCondition(ARGS *ap);
static	BOOL	SetSeparation(ARGS *ap);
static	BOOL	SetWidth(ARGS *ap);
static	BOOL	SetColumns(ARGS *ap);
static	BOOL	SetColumnWidth(ARGS *ap);
static	BOOL	SetPids(ARGS *ap);
static	BOOL	SetUsers(ARGS *ap);
static	BOOL	SetGroups(ARGS *ap);
static	BOOL	SetDefault(ARGS *ap);
static	BOOL	AddColumns(ARGS *ap);
static	BOOL	RemoveColumns(ARGS *ap);
static	void	DefaultOneOption(int id);
static	void	PrintUsage(void);
static	void	PrintVersion(void);


/*
 * Parse the command line options.
 * The program name argument must have already been removed.
 * Returns TRUE if they were successfully parsed.
 */
BOOL
ParseOptions(ARGS *ap, int depth)
{
	OPTION *	option;
	char *		str;
	BOOL		status;
	BOOL		success;

	success = TRUE;

	while (ap->count-- > 0) {
		str = *ap->table++;

		/*
		 * If this is not an option, then it must be the name
		 * of an option declaration, so expand that name and
		 * collect those options.
		 */
		if (*str != '-') {
			if (!ExpandOptionName(str, depth))
				return FALSE;

			continue;
		}

		/*
		 * It is a normal option, so look it up and handle it.
		 */
		str++;

		option = option_table;

		while (option->name && (strcmp(option->name, str) != 0))
			option++;

		status = TRUE;

		switch (option->id) {
			case OPT_SET_COLUMNS:
				status = SetColumns(ap);
				break;

			case OPT_ADD_COLUMNS:
				status = AddColumns(ap);
				break;

			case OPT_REMOVE_COLUMNS:
				status = RemoveColumns(ap);
				break;

			case OPT_SEPARATION:
				status = SetSeparation(ap);
				break;

			case OPT_WIDTH:
				status = SetWidth(ap);
				break;

			case OPT_SLEEP_TIME:
				status = SetSleepTime(ap);
				break;

			case OPT_ACTIVE_TIME:
				status = SetActiveTime(ap);
				break;

			case OPT_SYNC_TIME:
				status = SetSyncTime(ap);
				break;

			case OPT_COLUMN_WIDTH:
				status = SetColumnWidth(ap);
				break;

			case OPT_TOP_COUNT:
				status = SetTopCount(ap);
				break;

			case OPT_INIT_TIME:
				status = SetInitTime(ap);
				break;

			case OPT_CONDITION:
				status = SetCondition(ap);
				break;

			case OPT_NO_SELF:
				no_self = TRUE;
				break;

			case OPT_NO_ROOT:
				no_root = TRUE;
				break;

			case OPT_NO_HEADER:
				no_header = TRUE;
				break;

			case OPT_MY_PROCS:
				my_procs = TRUE;
				break;

			case OPT_ACTIVE_ONLY:
				active_only = TRUE;
				break;

			case OPT_CLEAR_SCREEN:
				clear_screen = TRUE;
				break;

			case OPT_READ_FILE:
				status = ReadFile(ap);
				break;

			case OPT_NORMAL_SORT:
				status = AppendColumnSort(ap, FALSE);
				break;

			case OPT_REVERSE_SORT:
				status = AppendColumnSort(ap, TRUE);
				break;

			case OPT_NORMAL_SORT_EXPR:
				status = AppendExpressionSort(ap, FALSE);
				break;

			case OPT_REVERSE_SORT_EXPR:
				status = AppendExpressionSort(ap, TRUE);
				break;

			case OPT_NO_SORT:
				ClearSorting();
				break;

			case OPT_DEFAULT:
				status = SetDefault(ap);
				break;

			case OPT_LOOP:
				loop = TRUE;
				break;

			case OPT_VERTICAL:
				vertical = TRUE;
				break;

			case OPT_PIDS:
				status = SetPids(ap);
				break;

			case OPT_USERS:
				status = SetUsers(ap);
				break;

			case OPT_GROUPS:
				status = SetGroups(ap);
				break;

			case OPT_HELP:
				PrintUsage();
				break;

			case OPT_LIST_COLUMNS:
				ListColumns();
				exit(0);

			case OPT_LIST_MACROS:
				ListMacros();
				exit(0);

			case OPT_VERSION:
				PrintVersion();
				exit(0);

			case OPT_END:
				break;

			case OPT_NO_INIT:
				fprintf(stderr,
"The -noinit option can only be used as the first command line argument\n");

				status = FALSE;
				break;

			default:
				fprintf(stderr, "Unknown option -%s\n", str);
				status = FALSE;
				break;
		}

		if (!status)
			success = FALSE;
	}

	return success;
}


/*
 * Expand an option definition.
 * Such a definition has to be complete within the expansion.
 * For the top level expansions only, we uppercase words for convenience.
 */
BOOL
ExpandOptionName(char *name, int depth)
{
	ARGS	args;
	char	upname[MAX_MACRO_LEN + 2];

	if (depth > MAX_OPTION_DEPTH) {
		fprintf(stderr, "Too many levels of option macros\n");

		return FALSE;
	}

	if (strlen(name) > MAX_MACRO_LEN) {
		fprintf(stderr, "Macro name \"%c%s\" is too long\n",
			*name + 'A' - 'a', name + 1);

		return FALSE;
	}

	/*
	 * If we are on the top level command line, then uppercase the
	 * macro name for convenience.
	 */
	if ((depth == 0) && islower(*name)) {
		strcpy(upname, name);

		name = upname;

		*name += 'A' - 'a';
	}

	if (islower(*name)) {
		fprintf(stderr, "Option macro \"%s\" must be upper case\n",
			name);

		return FALSE;
	}

	if (!ismacro(*name)) {
		fprintf(stderr, "String value \"%s\" is not a macro name\n",
			name);

		return FALSE;
	}

	if (!ExpandMacro(MACRO_TYPE_OPTION, name, &args))
		return FALSE;

	return ParseOptions(&args, depth + 1);
}


/*
 * Expand command line arguments for an option by looking for macro names
 * and expanding the macros recursively.  The expanded arguments are placed
 * into the specified table.  Returns the number of arguments resulting
 * from the expansions, or -1 on an error.
 */
int
ExpandArguments(ARGS *ap, char **table, int tablelen)
{
	char *	name;
	int	count;
	int	subcount;
	ARGS	args;

	count = 0;

	while ((ap->count > 0) && (**ap->table != '-')) {
		name = *ap->table++;
		ap->count--;

		/*
		 * Make sure there is room left in the table.
		 */
		if (tablelen <= 0) {
			fprintf(stderr, "Too many arguments\n");

			return -1;
		}

		/*
		 * If the argument is not a macro name, then just put it
		 * into the returned argument table 
		 */
		if (!ismacro(*name)) {
			*table++ = name;
			tablelen--;
			count++;

			continue;
		}

		/*
		 * The argument is a macro name, so expand it into more
		 * arguments.
		 */
		if (!ExpandMacro(MACRO_TYPE_COLUMN, name, &args))
			return -1;

		/*
		 * Add the result of the expansion into the table at this
		 * point, possibly doing a recursive expansion.
		 */
		subcount = ExpandArguments(&args, table, tablelen);

		if (subcount < 0)
			return -1;

		/*
		 * Increment past the number of entries that the expansion
		 * added to the table.
		 */
		table += subcount;
		tablelen -= subcount;
		count += subcount;
	}

	return count;
}


/*
 * Set the separation between columns.
 */
static BOOL
SetSeparation(ARGS *ap)
{
	char *	str;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing separation argument\n");

		return FALSE;
	}

	ap->count--;
	str = *ap->table++;

	separation = GetDecimalNumber(&str);

	if (*str || (separation < 0) || (separation > MAX_SEPARATION)) {
		fprintf(stderr, "Bad separation argument\n");

		return FALSE;
	}

	return TRUE;
}


/*
 * Read the specified file containing macro definitions.
 */
static BOOL
ReadFile(ARGS *ap)
{
	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing file argument\n");

		return FALSE;
	}

	ap->count--;

	return ParseFile(*ap->table++);
}


/*
 * Set the default value associated with one or more options.
 * If no option names are given, then all defaults are set.
 */
static BOOL
SetDefault(ARGS *ap)
{
	OPTION *option;
	char *	name;

	/*
	 * If no option name was given, then default everything.
	 */
	if ((ap->count <= 0) || (**ap->table == '-')) {
		DefaultAllOptions();

		return TRUE;
	}

	/*
	 * There was at least one option given, so default them.
	 */
	while ((ap->count > 0) && (**ap->table != '-')) {
		ap->count--;
		name = *ap->table++;

		option = option_table;

		while (option->name && (strcmp(option->name, name) != 0))
			option++;

		if (option->name == NULL) {
			fprintf(stderr, "Undefined option name \"%s\"\n", name);

			return FALSE;
		}

		DefaultOneOption(option->id);
	}

	return TRUE;
}


/*
 * Routine to default all option values.
 * This is called at program startup to initialize parameters,
 * and also when the user requests it.
 */
void
DefaultAllOptions(void)
{
	int	id;

	for (id = 0; id <= OPT_LAST_OPTION; id++)
		DefaultOneOption(id);
}


/*
 * Routine to default the specified option id.
 * This is only meaningful for certain options, but calling this for the
 * remaining options isn't a problem.  This routine never reports an error.
 */
static void
DefaultOneOption(int id)
{
	switch (id) {
		case OPT_SET_COLUMNS:
		case OPT_ADD_COLUMNS:
		case OPT_REMOVE_COLUMNS:
			DefaultColumns();
			break;

		case OPT_SEPARATION:
			separation = 2;
			break;

		case OPT_WIDTH:
			width = 80;
			break;

		case OPT_SLEEP_TIME:
			sleep_time = 10;
			break;

		case OPT_NO_SELF:
			no_self = FALSE;
			break;

		case OPT_MY_PROCS:
			my_procs = FALSE;
			break;

		case OPT_ACTIVE_ONLY:
			active_only = FALSE;
			break;

		case OPT_NO_ROOT:
			no_root = FALSE;
			break;

		case OPT_CLEAR_SCREEN:
			clear_screen = FALSE;
			break;

		case OPT_NORMAL_SORT:
		case OPT_REVERSE_SORT:
		case OPT_NORMAL_SORT_EXPR:
		case OPT_REVERSE_SORT_EXPR:
		case OPT_NO_SORT:
			ClearSorting();
			break;

		case OPT_LOOP:
			loop = FALSE;
			break;

		case OPT_VERTICAL:
			vertical = FALSE;
			break;

		case OPT_INIT_TIME:
			init_time = 1;
			break;

		case OPT_CONDITION:
			ClearCondition();
			break;

		case OPT_TOP_COUNT:
			top_count = 0;
			break;

		case OPT_NO_HEADER:
			no_header = FALSE;
			break;

		case OPT_PIDS:
			pid_count = 0;
			break;

		case OPT_USERS:
			user_count = 0;
			break;

		case OPT_GROUPS:
			group_count = 0;
			break;

		case OPT_ACTIVE_TIME:
			active_time = 60;
			break;

		case OPT_SYNC_TIME:
			sync_time = (60 * 5);
			break;

		case OPT_COLUMN_WIDTH:
			DefaultColumnWidths();
			break;

		default:
			break;
	}
}


/*
 * Set the condition for displaying of processes.
 */
static BOOL
SetCondition(ARGS *ap)
{
	char *	str;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing condition expression\n");

		return FALSE;
	}

	ap->count--;
	str = *ap->table++;

	return ParseCondition(str);
}


/*
 * Set the sleep time between updates.
 */
static BOOL
SetSleepTime(ARGS *ap)
{
	char *	str;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing sleep time argument\n");

		return FALSE;
	}

	ap->count--;
	str = *ap->table++;

	sleep_time = GetDecimalNumber(&str);

	if (*str || (sleep_time < 0)) {
		fprintf(stderr, "Bad sleep time argument\n");

		return FALSE;
	}

	return TRUE;
}


/*
 * Set the number of seconds to keep an active process in the display.
 */
static BOOL
SetActiveTime(ARGS *ap)
{
	char *	str;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing active time argument\n");

		return FALSE;
	}

	ap->count--;
	str = *ap->table++;

	active_time = GetDecimalNumber(&str);

	if (*str || (active_time < 0)) {
		fprintf(stderr, "Bad active time argument\n");

		return FALSE;
	}

	return TRUE;
}


/*
 * Set the number of seconds to force synchronization with process status.
 */
static BOOL
SetSyncTime(ARGS *ap)
{
	char *	str;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing sync time argument\n");

		return FALSE;
	}

	ap->count--;
	str = *ap->table++;

	sync_time = GetDecimalNumber(&str);

	if (*str || (sync_time < 0)) {
		fprintf(stderr, "Bad sync time argument\n");

		return FALSE;
	}

	return TRUE;
}


/*
 * Set the number of processes to show from the top.
 * The count can be omitted to indicate that we calculate it
 * from the terminal height.
 */
static BOOL
SetTopCount(ARGS *ap)
{
	char *	str;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		top_count = height - 1;

		if (!no_header)
			top_count--;

		if (top_count <= 0)
			top_count = 1;

		top_auto = TRUE;

		return TRUE;
	}

	ap->count--;
	str = *ap->table++;

	top_count = GetDecimalNumber(&str);

	if (*str || (top_count < 0)) {
		fprintf(stderr, "Bad top argument\n");

		return FALSE;
	}

	top_auto = FALSE;

	return TRUE;
}


/*
 * Set the list of pids to be examined.
 */
static BOOL
SetPids(ARGS *ap)
{
	char *	str;
	int	pid;
	int	i;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing pid argument\n");

		return FALSE;
	}

	while ((ap->count > 0) && (**ap->table != '-')) {
		ap->count--;
		str = *ap->table++;

		pid = GetDecimalNumber(&str);

		if (*str || (pid <= 0) || (pid > 30000)) {
			fprintf(stderr, "Bad pid argument\n");

			return FALSE;
		}

		for (i = 0; i < pid_count; i++) {
			if (pid_list[i] == pid)
				break;
		}

		if (i < pid_count)
			continue;

		if (pid_count >= MAX_PIDS) {
			fprintf(stderr, "Too many pids specified\n");

			return FALSE;
		}

		pid_list[pid_count++] = pid;
	}

	return TRUE;
}


/*
 * Set the list of users to be examined.
 */
static BOOL
SetUsers(ARGS *ap)
{
	char *	str;
	int	uid;
	int	i;

	CollectUserNames();

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing user argument\n");

		return FALSE;
	}

	while ((ap->count > 0) && (**ap->table != '-')) {
		ap->count--;
		str = *ap->table++;

		if (isdigit(*str)) {
			uid = GetDecimalNumber(&str);

			if (*str || (uid < 0)) {
				fprintf(stderr, "Bad uid argument\n");

				return FALSE;
			}
		} else {
			uid = FindUserId(str);

			if (uid < 0) {
				fprintf(stderr, "Unknown user \"%s\"\n", str);

				return FALSE;
			}
		}

		for (i = 0; i < user_count; i++) {
			if (user_list[i] == uid)
				break;
		}

		if (i < user_count)
			continue;

		if (user_count >= MAX_USERS) {
			fprintf(stderr, "Too many users specified\n");

			return FALSE;
		}

		user_list[user_count++] = uid;
	}

	return TRUE;
}


/*
 * Set the list of groups to be examined.
 */
static BOOL
SetGroups(ARGS *ap)
{
	char *	str;
	int	gid;
	int	i;

	CollectGroupNames();

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing group argument\n");

		return FALSE;
	}

	while ((ap->count > 0) && (**ap->table != '-')) {
		ap->count--;
		str = *ap->table++;

		if (isdigit(*str)) {
			gid = GetDecimalNumber(&str);

			if (*str || (gid < 0)) {
				fprintf(stderr, "Bad gid argument\n");

				return FALSE;
			}
		} else {
			gid = FindGroupId(str);

			if (gid < 0) {
				fprintf(stderr, "Unknown group \"%s\"\n", str);

				return FALSE;
			}
		}

		for (i = 0; i < group_count; i++) {
			if (group_list[i] == gid)
				break;
		}

		if (i < group_count)
			continue;

		if (group_count >= MAX_GROUPS) {
			fprintf(stderr, "Too many groups specified\n");

			return FALSE;
		}

		group_list[group_count++] = gid;
	}

	return TRUE;
}


/*
 * Set the sleep time before the initial update.
 */
static BOOL
SetInitTime(ARGS *ap)
{
	char *	str;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing init time argument\n");

		return FALSE;
	}

	ap->count--;
	str = *ap->table++;

	init_time = GetDecimalNumber(&str);

	if (*str || (init_time < 0)) {
		fprintf(stderr, "Bad init time argument\n");

		return FALSE;
	}

	return TRUE;
}


/*
 * Set the display width.
 */
static BOOL
SetWidth(ARGS *ap)
{
	char *	str;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing width argument\n");

		return FALSE;
	}

	ap->count--;
	str = *ap->table++;

	width = GetDecimalNumber(&str);

	if (*str || (width <= 0)) {
		fprintf(stderr, "Bad width argument\n");

		return FALSE;
	}

	if (width > MAX_WIDTH)
		width = MAX_WIDTH;

	return TRUE;
}


/*
 * Set the width of the specified column to the indicated value,
 * or else back to its default value if no value was given.
 */
static BOOL
SetColumnWidth(ARGS *ap)
{
	COLUMN *column;
	char *	name;
	char *	str;
	int	width;

	if ((ap->count <= 0) || (**ap->table == '-')) {
		fprintf(stderr, "Missing Column name\n");

		return FALSE;
	}

	ap->count--;
	name = *ap->table++;

	/*
	 * Find the column from the name.
	 */
	column = FindColumn(name);

	if (column == NULL) {
		fprintf(stderr, "Bad column name %s\n", name);

		return FALSE;
	}

	/*
	 * If no additional argument was given,
	 * then set the column's width back to its initial value.
	 */
	if ((ap->count <= 0) || (**ap->table == '-')) {
		column->width = column->initwidth;

		return TRUE;
	}

	/*
	 * There was a width specified, so parse it and set the
	 * column width to that value.
	 */
	ap->count--;
	str = *ap->table++;

	width = GetDecimalNumber(&str);

	if (*str || (width <= 0)) {
		fprintf(stderr, "Bad column width argument\n");

		return FALSE;
	}

	if (width > MAX_WIDTH)
		width = MAX_WIDTH;

	column->width = width;

	return TRUE;
}


/*
 * Set the columns to display according to the indicated names.
 */
static BOOL
SetColumns(ARGS *ap)
{
	show_count = 0;

	return AddColumns(ap);
}


/*
 * Add the columns to display according to the indicated names.
 * The columns are added to the end of the list of columns to display,
 * or else at the column number specified before the column name.
 * This removes any previous usage of the column.
 */
static BOOL
AddColumns(ARGS *ap)
{
	char *	name;
	COLUMN *column;
	int	count;
	int	i;
	int	j;
	int	pos;
	char *	table[MAX_WORDS];

	count = ExpandArguments(ap, table, MAX_WORDS);

	if (count < 0)
		return FALSE;

	/*
	 * Default the insert position to be after all existing columns.
	 */
	pos = show_count;

	/*
	 * Go through the column names one by one, inserting them into
	 * the show list.
	 */
	for (i = 0; i < count; i++) {
		name = table[i];

		/*
		 * If the column name is actually a number, then the
		 * argument is the column number to insert columns at.
		 * Subtract one to convert it to an array offset.
		 */
		if (isdigit(*name)) {
			pos = 0;

			while (isdigit(*name))
				pos = pos * 10 + *name++ - '0';

			if (*name || (pos < 0)) {
				fprintf(stderr, "Bad column position argument\n");

				return FALSE;
			}

			pos--;

			continue;
		}

		/*
		 * This is actually a column name, so look it up.
		 */
		column = FindColumn(name);

		if (column == NULL) {
			fprintf(stderr, "Bad column name %s\n", name);

			return FALSE;
		}

		/*
		 * Remove the column if it was already in the column list.
		 */
		for (j = 0; j < show_count; j++) {
			if (column == show_list[j]) {
				while (j < show_count) {
					show_list[j] = show_list[j + 1];
					j++;
				}

				show_count--;
			}
		}

		/*
		 * Now insert the column at the specified column position.
		 * Do this by moving all columns at the position up one,
		 * and putting the new column at the proper location.
		 * Make sure the position is valid beforehand, and increment
		 * the position afterward so that the next column to be
		 * inserted goes in after this one.
		 */
		if (pos < 0)
			pos = 0;

		if (pos > show_count)
			pos = show_count;

		for (j = show_count; j > pos; j--)
			show_list[j] = show_list[j - 1];

		show_list[pos++] = column;
		show_list[++show_count] = NULL;
	}

	return TRUE;
}


/*
 * Remove the indicated columns from the display.
 */
static BOOL
RemoveColumns(ARGS *ap)
{
	char *	name;
	COLUMN *column;
	int	count;
	int	i;
	int	j;
	char *	table[MAX_WORDS];

	count = ExpandArguments(ap, table, MAX_WORDS);

	if (count < 0)
		return FALSE;

	for (i = 0; i < count; i++) {
		name = table[i];

		column = FindColumn(name);

		if (column == NULL) {
			fprintf(stderr, "Bad column name %s\n", name);

			return FALSE;
		}

		for (j = 0; j < show_count; j++) {
			if (column == show_list[j]) {
				while (j < show_count) {
					show_list[j] = show_list[j + 1];
					j++;
				}

				show_count--;
			}
		}
	}

	return TRUE;
}


/*
 * Usage text
 */
static	char *	usage_text[] =
{
"Usage: ips <show-options> <condition-options> <other-options> <macro-names>",
"",
"<show-options> specify how to display the output:",
"  -col <col> ...           Set columns as the display list",
"  -addcol <col> ...        Add columns to display list",
"  -remcol <col> ...        Remove columns from display list",
"  -sort <col> ...          Add columns to sorting list as normal sort",
"  -revsort <col> ...       Add columns to sorting list as reverse sort",
"  -sortexpr <expr>         Add expression to sorting list as normal sort",
"  -revsortexpr <expr>      Add expression to sorting list as reverse sort",
"  -nosort                  Clear the sorting list",
"  -sep <count>             Set the separation between columns",
"  -width <size>            Set the total width of output for all columns",
"  -colwidth <col> [width]  Set (or default) the width for one column",
"  -clear                   Clear the screen before outputting",
"  -noheader                Suppress display of header line",
"  -vert                    Output status in vertical format (multi-line)",
"",
"<condition-options> restrict which processes can be displayed:",
"  -noself                  Do not show the ips process itself",
"  -noroot                  Do not show processes owned by root",
"  -my                      Show processes with my own user id",
"  -pid <pid> ...           Show specified process ids",
"  -user <user> ...         Show specified user names or ids",
"  -group <group> ...       Show specified group names or ids",
"  -active                  Show active processes",
"  -activetime <seconds>    Set number of seconds processes remain active",
"  -top [count]             Show number of processes from the top or what fits",
"  -cond <expr>             Show processes for which the expression is nonzero",
"",
"<other-options> are:",
"  -loop                    Loop forever displaying the process list",
"  -sleep <seconds>         Set the sleep interval between loops",
"  -initsleep <seconds>     Set the initial sleep time for active checks",
"  -synctime <seconds>      Synchronize with some status at least this often",
"  -read <filename>         Read definitions from the specified filename",
"  -noinit                  Don't read \"~/.ipsrc\" (must be first option)",
"  -end                     Used to end argument lists if needed",
"  -help                    Display this message and exit",
"  -listcolumns             Display the list of allowable columns and exit",
"  -listmacros              Display the list of defined macros and exit",
"  -version                 Display version information and exit",
"  -default [name] ...      Set specified options back to their default values",
"",
"<macro-names> are a list of macros (defined in the initialization files)",
"which are to be expanded into options.  The first character of each macro",
"name is converted to upper case for convenience.",
NULL
};


static void
PrintUsage(void)
{
	char **	text;

	for (text = usage_text; *text; text++)
		fprintf(stderr, "%s\n", *text);

	exit(1);
}


static void
PrintVersion(void)
{
	printf("Version:   %s\n", VERSION);
	printf("Built on:  %s\n", __DATE__);
	printf("Author:    %s\n", AUTHOR_NAME);
	printf("Email:     %s\n", AUTHOR_EMAIL);
}

/* END CODE */
